import uuid
from datetime import datetime
from time import sleep
from typing import Union

from dateutil import parser
from dateutil.parser import ParserError
from pytz import timezone

from .general_runtime_vars import env_vars


def generate_guid():
    return str(uuid.uuid1())


def parse_headers(msg_header):
    """
    Converts headers to a dict
    :param msg_header: A message .headers() (confluent) or .headers (faust) instance
    :return: a decoded dict version of the headers
    """
    if msg_header:
        msg_header = dict(msg_header)
        return {key: value.decode() for key, value in msg_header.items()}
    return {}


def log_and_raise_error(metrics_manager, error):
    """
    Since the metric manager pushing is a separate thread, ensure an exception gets sent to prometheus
    """
    metrics_manager.inc_message_errors(error)
    sleep(int(env_vars()['METRICS_PUSH_RATE']) * 2 + 1)
    raise


def universal_datetime_converter(incoming_datetime: Union[str, datetime], incoming_timezone: timezone = timezone("Etc/UTC")) -> str:
    """
    Converts datetime values of all kinds into a normalized format
    Assumes midnight if datetime does not include time
    """
    universal_format = '%Y-%m-%dT%H:%M:%SZ'

    try:
        datetime_to_convert = parser.parse(str(incoming_datetime))
    except ParserError:
        try:
            datetime_to_convert = datetime.utcfromtimestamp(int(incoming_datetime))
        except Exception:
            return ""

    if datetime_to_convert.tzinfo is None or datetime_to_convert.tzinfo.utcoffset(datetime_to_convert) is None:
        datetime_to_convert = incoming_timezone.localize(datetime_to_convert)

    with_offset_for_given_timezone = incoming_timezone.normalize(datetime_to_convert)
    datetime_to_convert = with_offset_for_given_timezone.astimezone(timezone("Etc/UTC"))

    converted_datetime = datetime_to_convert.strftime(universal_format)
    return converted_datetime



# runtime environment variables used by both confluent-kafka and faust implementations

from os import environ
from .env_var_generator import env_vars_creator


def default_env_vars():
    """
    Environment variables that have defaults if not specified.
    """
    return {
        'LOGLEVEL': environ.get('LOGLEVEL', 'INFO'),
        'KAFKA_CLUSTER': environ.get('KAFKA_CLUSTER', 'bifrost-kafka-bootstrap:9093'),
        'SCHEMA_REGISTRY_URL': environ.get('SCHEMA_REGISTRY_URL', 'http://bifrost-schema-registry:8081'),
        'USE_SSL': environ.get('USE_SSL', 'true'),
        'SSL_CA_LOCATION': environ.get('SSL_CA_LOCATION', '/opt/app-root/cert/client-trusted-certs/trusted-certs.pem'),
        'SSL_CERTIFICATE_LOCATION': environ.get('SSL_CERTIFICATE_LOCATION', '/opt/app-root/cert/clients-ca-cert/ca.crt'),
        'SSL_KEY_LOCATION': environ.get('SSL_KEY_LOCATION', '/opt/app-root/cert/clients-ca/ca.key'),
        'USE_SASL': environ.get('USE_SASL', 'false'),
        'SASL_USERNAME': environ.get('SASL_USERNAME', ''),
        'SASL_PASSWORD': environ.get('SASL_PASSWORD', ''),

        # Metrics Manager
        'DO_METRICS_PUSHING': environ.get('DO_METRICS_PUSHING', 'true'),
        'METRICS_PUSH_RATE': environ.get('METRICS_PUSH_RATE', '10'),
        'METRICS_SERVICE_NAME': environ.get('METRICS_SERVICE_NAME', 'bifrost-metrics-cache-headless.mktg-ops--kafka.svc.cluster.local'),
        'METRICS_SERVICE_PORT': environ.get('METRICS_SERVICE_PORT', '8080'),
        'METRICS_POD_PORT': environ.get('METRICS_POD_PORT', '9091'),

        'CONSUMER_HEARTBEAT_TIMEOUT_SECONDS': '90'}


def required_env_vars():
    """
    Environment variables that require a value (aka no default specified).
    """
    return {
        'HOSTNAME': environ['HOSTNAME'],  # NOTE: Every Openshift pod has a default HOSTNAME (its own pod name).
        'APP_NAME': environ['APP_NAME']}


def all_env_vars():
    return {
        **default_env_vars(),
        **required_env_vars()
    }


env_vars = env_vars_creator(all_env_vars)



import setuptools

with open("README.md", "r") as file_obj:
    long_description = file_obj.read()

install_requires = [
    "prometheus_client",
    "requests",
    "pytz",
    "python-dateutil"
]

faust_requires = [
    "cython==0.29.17",
    "nubium-faust[rocksdb]",
    "nubium-aiokafka",
    "idna<3",
    "chardet<4",
    "python-schema-registry-client>1.8.1",
]

dev_requires = faust_requires + [
    "pytest",
    "twine",
]

packages = setuptools.find_packages()

setuptools.setup(
    name="nubium-utils",
    version="0.25.4",
    author="Edward Brennan",
    author_email="ebrennan@redhat.com",
    description="Some Kafka utility functions and patterns for the nubium project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.corp.redhat.com/mkt-ops-de/nubium-utils.git",
    packages=packages,
    install_requires=install_requires,
    extras_require={"dev": dev_requires, "faust": faust_requires},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



def env_vars_creator(dict_func):
    """
    Allows you to call a dict without constantly re-instantiating it or assigning it to a constant.
    The dict is not created until you evaluate the return of this function (another function) for the first time.
    Every subsequent evaluation of the returned function will return the previously created dict (it's just getting
    the same dict via an infinite generator).

    This allows you to avoid having a dict constant, which would cause import failures at init time if you are missing
    any of the required environment variables. This is mostly relevant with regards to unit tests, where some env vars
    may not be defined until during/after setup.
    """
    def get_generator(dict_func):
        d = dict_func()
        while True:
            yield d

    def get_dict(dict_func):
        gen = get_generator(dict_func)
        return lambda: next(gen)

    return get_dict(dict_func)



"""
A multithreading monitoring setup for use with non-streams Kafka applications

The thread target function calls the metrics pusher every couple of seconds,
and then waits.
"""
import threading
import time

from .metrics_manager import MetricsManager


def metrics_push_function(metrics_manager, push_rate: int):
    """
    Sets gateways and pushes metrics

    Used as a multi-threading target
    :param metrics_manager: (MetricsManager) Pushes prometheus metrics to gateway
    :param push_rate: (int) Amount of time to sleep between pushes
    """
    while True:
        time.sleep(push_rate)
        metrics_manager.push_metrics()


def get_metrics_pushing_thread(metrics_manager, push_rate: int):
    """
    Creates and returns threading object for monitoring the app

    Thread is a daemon since it should exit as soon as the regular program exits
    :param metrics_manager: (MetricsManager) Pushes prometheus metrics to gateway
    :param push_rate: (int) Amount of time to sleep between pushes
    """
    return threading.Thread(target=metrics_push_function, args=(metrics_manager, push_rate), daemon=True)


def start_pushing_metrics(metrics_manager: MetricsManager, push_rate: int):
    """
    Creates and starts a monitoring thread

    :param metrics_manager: (MetricsManager) Pushes prometheus metrics to gateway
    :param push_rate: (int) Amount of time to sleep between pushes
    """
    monitoring_thread = get_metrics_pushing_thread(metrics_manager, push_rate)
    monitoring_thread.start()
